"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DuckDBDriver = void 0;
const base_driver_1 = require("@cubejs-backend/base-driver");
const shared_1 = require("@cubejs-backend/shared");
const util_1 = require("util");
const stream = __importStar(require("stream"));
const duckdb_1 = require("duckdb");
const DuckDBQuery_1 = require("./DuckDBQuery");
const HydrationStream_1 = require("./HydrationStream");
const { version } = require('../../package.json');
const DuckDBToGenericType = {
    // DATE_TRUNC returns DATE, but Cube Store still doesn't support DATE type
    // DuckDB's driver transform date/timestamp to Date object, but HydrationStream converts any Date object to ISO timestamp
    // That's why It's safe to use timestamp here
    date: 'timestamp',
};
class DuckDBDriver extends base_driver_1.BaseDriver {
    constructor(config = {}) {
        super();
        this.config = config;
        this.initPromise = null;
        this.schema = this.config.schema || (0, shared_1.getEnv)('duckdbSchema', this.config);
    }
    toGenericType(columnType) {
        if (columnType.toLowerCase() in DuckDBToGenericType) {
            return DuckDBToGenericType[columnType.toLowerCase()];
        }
        return super.toGenericType(columnType.toLowerCase());
    }
    async init() {
        const token = (0, shared_1.getEnv)('duckdbMotherDuckToken', this.config);
        const dbPath = (0, shared_1.getEnv)('duckdbDatabasePath', this.config);
        // Determine the database URL based on the provided db_path or token
        let dbUrl;
        if (dbPath) {
            dbUrl = dbPath;
        }
        else if (token) {
            dbUrl = `md:?motherduck_token=${token}&custom_user_agent=Cube/${version}`;
        }
        else {
            dbUrl = ':memory:';
        }
        let dbOptions;
        if (token) {
            dbOptions = { custom_user_agent: `Cube/${version}` };
        }
        // Create a new Database instance with the determined URL and custom user agent
        const db = new duckdb_1.Database(dbUrl, dbOptions);
        // Under the hood all methods of Database uses internal default connection, but there is no way to expose it
        const defaultConnection = db.connect();
        const execAsync = (0, util_1.promisify)(defaultConnection.exec).bind(defaultConnection);
        try {
            await execAsync('INSTALL httpfs');
        }
        catch (e) {
            if (this.logger) {
                console.error('DuckDB - error on httpfs installation', {
                    e
                });
            }
            // DuckDB will lose connection_ref on connection on error, this will lead to broken connection object
            throw e;
        }
        try {
            await execAsync('LOAD httpfs');
        }
        catch (e) {
            if (this.logger) {
                console.error('DuckDB - error on loading httpfs', {
                    e
                });
            }
            // DuckDB will lose connection_ref on connection on error, this will lead to broken connection object
            throw e;
        }
        const configuration = [
            {
                key: 's3_region',
                value: (0, shared_1.getEnv)('duckdbS3Region', this.config),
            },
            {
                key: 's3_endpoint',
                value: (0, shared_1.getEnv)('duckdbS3Endpoint', this.config),
            },
            {
                key: 's3_access_key_id',
                value: (0, shared_1.getEnv)('duckdbS3AccessKeyId', this.config),
            },
            {
                key: 's3_secret_access_key',
                value: (0, shared_1.getEnv)('duckdbS3SecretAccessKeyId', this.config),
            },
            {
                key: 'memory_limit',
                value: (0, shared_1.getEnv)('duckdbMemoryLimit', this.config),
            },
            {
                key: 'schema',
                value: (0, shared_1.getEnv)('duckdbSchema', this.config),
            },
            {
                key: 's3_use_ssl',
                value: (0, shared_1.getEnv)('duckdbS3UseSsl', this.config),
            },
            {
                key: 's3_url_style',
                value: (0, shared_1.getEnv)('duckdbS3UrlStyle', this.config),
            },
            {
                key: 's3_session_token',
                value: (0, shared_1.getEnv)('duckdbS3SessionToken', this.config),
            }
        ];
        for (const { key, value } of configuration) {
            if (value) {
                try {
                    await execAsync(`SET ${key}='${value}'`);
                }
                catch (e) {
                    if (this.logger) {
                        console.error(`DuckDB - error on configuration, key: ${key}`, {
                            e
                        });
                    }
                }
            }
        }
        if (this.config.initSql) {
            try {
                await execAsync(this.config.initSql);
            }
            catch (e) {
                if (this.logger) {
                    console.error('DuckDB - error on init sql (skipping)', {
                        e
                    });
                }
            }
        }
        return {
            defaultConnection,
            db
        };
    }
    informationSchemaQuery() {
        if (this.schema) {
            return `${super.informationSchemaQuery()} AND table_catalog = '${this.schema}'`;
        }
        return super.informationSchemaQuery();
    }
    getSchemasQuery() {
        if (this.schema) {
            return `
        SELECT table_schema as ${super.quoteIdentifier('schema_name')}
        FROM information_schema.tables
        WHERE table_catalog = '${this.schema}'
        GROUP BY table_schema
      `;
        }
        return super.getSchemasQuery();
    }
    async getInitiatedState() {
        if (!this.initPromise) {
            this.initPromise = this.init();
        }
        try {
            return await this.initPromise;
        }
        catch (e) {
            this.initPromise = null;
            throw e;
        }
    }
    static dialectClass() {
        return DuckDBQuery_1.DuckDBQuery;
    }
    async query(query, values = [], _options) {
        const { defaultConnection } = await this.getInitiatedState();
        const fetchAsync = (0, util_1.promisify)(defaultConnection.all).bind(defaultConnection);
        const result = await fetchAsync(query, ...values);
        return result.map((item) => {
            (0, HydrationStream_1.transformRow)(item);
            return item;
        });
    }
    async stream(query, values, { highWaterMark }) {
        const { db } = await this.getInitiatedState();
        // new connection, because stream can break with
        // Attempting to execute an unsuccessful or closed pending query result
        // PreAggregation queue has a concurrency limit, it's why pool is not needed here
        const connection = db.connect();
        const closeAsync = (0, util_1.promisify)(connection.close).bind(connection);
        try {
            const asyncIterator = connection.stream(query, ...(values || []));
            const rowStream = stream.Readable.from(asyncIterator, { highWaterMark }).pipe(new HydrationStream_1.HydrationStream());
            return {
                rowStream,
                release: async () => {
                    await closeAsync();
                }
            };
        }
        catch (e) {
            await closeAsync();
            throw e;
        }
    }
    async testConnection() {
        await this.query('SELECT 1', []);
    }
    readOnly() {
        return false;
    }
    async release() {
        if (this.initPromise) {
            const { db } = await this.initPromise;
            const close = (0, util_1.promisify)(db.close).bind(db);
            this.initPromise = null;
            await close();
        }
    }
}
exports.DuckDBDriver = DuckDBDriver;
//# sourceMappingURL=DuckDBDriver.js.map